from pycel import *
from apricotpc import ApricotPropertyClass, NoneParBlock
import math

class SensorEnvironment(ApricotPropertyClass):
    """
    cel property class
    Sending beam to get useful information about entities and geometry.
    """

    def __init__(self, oreg):
        ApricotPropertyClass.__init__(self, oreg)
        self.SubscribeMessage("cel.timer.sensor.environment", self.Sensor)
        #: cel property type:vector3 rw
        #: Sampling beam start (relative to owner's origin in local coordinates)
        self._start = csVector3(0.0, 0.0, 0.0)
        self._start.z = -self._start.z
        #: cel property type:vector3 rw
        #: Sampling beam end (relative to start and normalized direction vector in local coordinates)
        self._end = csVector3(0.0, 0.0, 1.0)
        self._end.z = -self._end.z
        #: cel property type:bool r
        #: Character on ground True/False
        self.onground = False
        # Received information:
        # Detected iCelEntity or 0
        self.target_entity = 0
        # Detected iMeshWrapper or 0
        self.target_mesh = 0
        self.distance_mesh = 0.0
        self.cdsys = oreg.Get(iCollideSystem)

    def GetStart(self):
        return self._start

    def SetStart(self,val):
        self._start = val
        self._start.z = -self._start.z

    def GetEnd(self):
        return self._end

    def SetEnd(self,val):
        self._end = val
        self._end.z = -self._end.z

    end = property(GetEnd,SetEnd)
    start = property(GetStart,SetStart)

    def StartTick(waittime,func,repeat=False,step=CEL_EVENT_PRE,):
        PhysicalLayer.CallbackOnce(self,waittime,step)

    def SetEntity(self, entity):
        ApricotPropertyClass.SetEntity(self, entity)
        if not entity: return
        self.timer = celTimer(entity)
        self.mesh = celGetMesh(entity)
        self.meshwrapper = self.mesh.Mesh
        self.movable = self.meshwrapper.GetMovable()
        self.linmove = celLinearMovement(entity)
        self.onground = self.linmove.IsOnGround()
        self.sector = self.movable.GetSectors()[0]
        self.timer.WakeUp(100, True, "sensor.environment")

    def Sensor(self, args):
        # Prepare position of the beam
        self.sector = self.movable.GetSectors()[0]
        self.mesh.Mesh.SetFlagsRecursive(CS_ENTITY_NOHITBEAM)
        # entities kickable, carryable, climbable
        self.SensorEntity(csVector3(0.0, 0.29, 0.0), csVector3(0.0, 0.0, -1.0))
        # distance from ground gliding and edges
        self.SensorMesh(csVector3(0.0, 0.29, 0.25), csVector3(0.0, -100.0, 0.0))
        self.SensorGround()
        self.mesh.Mesh.SetFlagsRecursive(CS_ENTITY_NOHITBEAM, 0)

    def SensorEntity(self, bstart, bend):
        # Setting start and end vector
        beam_start = self.movable.GetFullPosition() + bstart
        beam_end = beam_start + (self.movable.GetFullTransform().This2Other(bend + bstart) - self.movable.GetFullTransform().This2Other(bstart))
        # Detecting entities
        entlist = PhysicalLayer.FindNearbyEntities(self.sector, beam_start, beam_end, False)
        if len(entlist):
            self.target_entity = entlist[0]
        else:
            self.target_entity = 0

    def SensorMesh(self, bstart, bend):
        # Setting start and end vector
        beam_start = self.movable.GetFullPosition() + bstart
        beam_end = beam_start + (self.movable.GetFullTransform().This2Other(bend + bstart) - self.movable.GetFullTransform().This2Other(bstart))
        # Visualizing
        arrow = PhysicalLayer.FindEntity("arrow")
        if arrow:
            arrow_mesh = celGetMesh(arrow)
            if arrow_mesh:
                arrow_mesh.MoveMesh(self.sector, beam_end)
        # Detecing meshes
        #hitbeam = self.meshwrapper.HitBeam(beam_start, beam_end)
        #print "Hit from", round(beam_start.x, 2), round(beam_start.y, 2), round(beam_start.z, 2), "to", round(beam_end.x, 2), round(beam_end.y, 2), round(beam_end.z, 2), "=", hitbeam.hit
        hitbeam = csColliderHelper.TraceBeam(self.cdsys, self.sector, beam_start, beam_end, False)
        if hitbeam:
            if hitbeam.sqdistance > 0.0:
                self.distance_mesh = math.sqrt(hitbeam.sqdistance)
        #print "HIT DISTANCE=", self.distance_mesh
        #print "closest mesh:", hitbeam.closest_mesh

    def SensorGround(self):
        # Detecting ground / fly
        if self.onground == False and self.linmove.IsOnGround() == True:
            self.SendMessage("apricot.sensor.environment.onground")
            self.onground = True
        elif self.onground == True and self.linmove.IsOnGround() == False:
            self.SendMessage("apricot.sensor.environment.inair")
            self.onground = False

CEL_IMPLEMENT_FACTORY(SensorEnvironment,"apricot.sensor.environment")
