from pycel import Clock,pyPcCommon,ObjectRegistry,PhysicalLayer,CEL_EVENT_PRE
import random

class ActionBase(object):
    """
    Base class for timed actions. Sets initial time the first time it is run
    (which can be time after the action is actually created).
    """
    def Do(self):
        self.starttime = Clock.CurrentTicks
        self.do()
        self.Do = self.do

class FadedAction(ActionBase):
    """
    Fade a value between min and max, in duration time
    """
    def __init__(self,func,start,end,duration):
        """
        @par func: function to execute to change the value
        @par start: starting value
        @par end: ending value
        @par duration: total duration for the action
        """
        self.func = func
        self.start = start
        self.end = end
        self.diff = self.end-self.start
        self.duration = float(duration)
    def do(self):
        elapsed = Clock.CurrentTicks - self.starttime
        if elapsed > self.duration:
            self.func(self.end)
            return True
        self.func(((float(elapsed)/self.duration)*self.diff)+self.start)

class TimedAction(ActionBase):
    """
    Run an action in a given time
    """
    def __init__(self,func,action_time):
        """
        @par func: function to execute when time expires
        @par action_time: time until the function executes
        """
        self.action_time = action_time
        self.func = func
    def do(self):
        elapsed = Clock.CurrentTicks - self.starttime
        if elapsed > self.action_time:
            self.func()
            return True

class MixerAction(ActionBase):
    """
    Action to mix several weighted functions, based on an assigned
    value to each of them. Will linearly interpolate two closest values
    to a value got from an attribute in an object.
    """
    def __init__(self,funcs,obj,prop,duration=None):
        """
        @par funcs: functions that receive the weights
        @par obj: object to get the syncing value from
        @par prop: property on which the syncing will be done
        @par duration: duration for the action
        """
        self.funcs = map(lambda s:s[0],funcs)
        self.vals = map(lambda s:s[1],funcs)
        self.obj = obj
        self.prop = prop
        self.duration=duration
    def do(self):
        elapsed = Clock.CurrentTicks - self.starttime
        if self.duration and elapsed > self.duration:
            return True
        val = getattr(self.obj,self.prop)
        if val < self.vals[0]:
            self.funcs[0](1.0)
            for func in self.funcs[1:]:
                func(0.0)
        elif val > self.vals[-1]:
            self.funcs[-1](1.0)
            for func in self.funcs[:-1]:
                func(0.0)
        else:
           idx = 1
           while val > self.vals[idx]:
                idx+=1
           val1 = self.vals[idx-1]
           val2 = self.vals[idx]
           dist = float(val2 - val1)
           curr_val = float(val - val1)
           self.funcs[idx-1](curr_val/dist)
           self.funcs[idx](1-(curr_val/dist))
           for func in self.funcs[:idx-1]+self.funcs[idx+1:]:
               func(0.0)

class RandomAction(ActionBase):
    """
    Action to randomly play a number of functions.
    """
    def __init__(self,funcs,min_time,max_time,duration=None):
        """
        @par funcs: functions to randomize
        @par min_time: minimum time between calls
        @par max_time: maximum time between calls
        @par duration: total duration for the action
        """
        self.funcs = funcs
        self.min_time = min_time
        self.max_time = max_time
        self.action_time = -1
        self.duration = duration
    def do(self):
        elapsed = Clock.CurrentTicks - self.starttime
        if self.duration and elapsed > self.duration:
            return True
        if elapsed > self.action_time:
            nextaction = random.randint(0,len(self.funcs)-1)
            self.funcs[nextaction]()
            self.action_time = elapsed+random.randint(self.min_time,self.max_time)

class ActionSequencer(object):
    """
    A sequencer to run a number of actions one after the other.
    """
    def __init__(self):
        self.actions = []
        self.autodestroy = False
    def append(self,action):
       """
       Add an action to the sequencer.
       """
       self.actions.append(action)
    def Do(self):
        if not self.actions:
            if self.autodestroy:
                return True
            return
        if self.actions[0].Do():
            self.actions.pop(0)

class TimedActionsManager(pyPcCommon):
    """
    A manager to run actions every frame.
    """
    def __init__(self): # XXX not a property class!
       pyPcCommon.__init__(self,ObjectRegistry)
       self.actions = {}
    def append(self,action,section=None):
       """
       Add an action to the manager.
       """
       PhysicalLayer.CallbackEveryFrame(self,CEL_EVENT_PRE)
       self._append(action,section)
       self._oldappend = self.append
       self.append = self._append
       return 0
    def clearall(self):
       """
       Clear all actions.
       """
       self.actions = {}
       self.append = self._oldappend
    def clear(self,section=None):
       if section in self.actions:
           self.actions.pop(section)
    def _append(self,action,section=None):
       self.actions[section] = action
       return len(self.actions)-1
    def TickEveryFrame(self):
       self.Do()
    def Do(self):
       finished = []
       for section,action in self.actions.items():
           if action.Do():
               finished.append(section)
       for section in finished:
           self.actions.pop(section)
       if not self.actions:
           self.clearall()

# Global manager
ActionsManager = TimedActionsManager()


def test(self):
    def func(val):
        print val

    def done():
        print "DONE"

    seq = ActionSequencer()
    seq.append(FadedAction(func,0.0,1000.0,10000))
    seq.append(FadedAction(func,1000.0,0.0,10000))
    seq.append(FadedAction(func,csVector3(0),csVector3(10.0),1000))
    seq.append(TimedAction(done,1000))

    ActionsManager.append(seq)

